from gradio_client import Client
from typing import Any
from gradio_client import utils as client_utils
import gradio as gr
from gradio.routes import App as GR_APP
import asyncio
from .gr_types.gr_type_normalizer import dereference_gradio_api
from .gr_types.gr_handle_file import (
    normalize_parameters_schema,
    normalize_returns_schema,
    process_inputs,
    extract_filedata_paths,
)
from fastapi import APIRouter, Request
from .utils.load_remote_ui import from_spaces_blocks_local_storage


def gr_client_wrapper(client: Client, api_name: str):
    config = (
        client.view_api(
            print_info=False,
            return_format="dict"
        )["named_endpoints"][api_name]
    )
    schema = dereference_gradio_api(config)

    async def core_gradio_client_infrence(
        request: Request
    ):
        data = await request.json()
        processed_data = process_inputs(
            data, paths=extract_filedata_paths(schema["parameters"]))

        job = client.submit(**processed_data, api_name=api_name)
        while not job.done():
            await asyncio.sleep(0.01)

        result = job.result()
        if len(schema["returns"]["properties"]) == 1:
            result = [result]
        ret = dict()
        for key, value in zip(schema["returns"]["properties"].keys(), result):
            ret[key] = value
        return ret

    return core_gradio_client_infrence


def gr_filedata_paths_wrapper(client: Client, api_name: str):
    config = (
        client.view_api(
            print_info=False,
            return_format="dict"
        )["named_endpoints"][api_name]
    )
    schema = dereference_gradio_api(config)
    inputs_paths = extract_filedata_paths(schema["parameters"])
    outputs_paths = extract_filedata_paths(schema["returns"])

    async def get_paths():
        return {
            "inputs": inputs_paths,
            "outputs": outputs_paths
        }

    return get_paths


def gr_original_schema_wrapper(client: Client, api_name: str):
    config = (
        client.view_api(
            print_info=False,
            return_format="dict"
        )["named_endpoints"][api_name]
    )
    schema = dereference_gradio_api(config)

    async def get_paths():
        return schema

    return get_paths


class RemoteGradioApplication:
    src: str
    blocks: gr.Blocks

    client: Client
    client_kwargs: dict
    client_document: str
    info_dict: dict

    router: APIRouter

    def __init__(
        self,
        src: str,
        client_kwargs: dict | None = None,
        router_kwargs: dict | None = None,
    ):
        self.src = src
        self.client_kwargs = client_kwargs or {}
        self.router_kwargs = router_kwargs or {}

        self.__init_gr_client()
        self.__init_api_router()

    def __init_gr_client(self):
        self.client = Client(self.src, **self.client_kwargs)
        self.client.view_api()  # print information
        self.client_document = self.client.view_api(
            print_info=False,
            return_format="str",
        )
        self.info_dict = self.client.view_api(
            print_info=False,
            return_format="dict",
        )

    def __init_api_router(self):
        router = self.router = APIRouter(**self.router_kwargs)
        for api_name, config_dict in self.info_dict["named_endpoints"].items():
            schema = dereference_gradio_api(config_dict)
            normalized_schema = dict()
            normalized_schema["parameters"] = normalize_parameters_schema(
                schema["parameters"])
            normalized_schema["returns"] = normalize_returns_schema(
                schema["returns"])

            router.post(
                path=api_name,
                openapi_extra={
                    "requestBody": {
                        "content": {
                            "application/json": {
                                "schema": normalized_schema["parameters"]
                            }
                        }
                    },
                    "responses": {
                        "200": {
                            "description": "Successful response",
                            "content": {
                                "application/json": {
                                    "schema": normalized_schema["returns"]
                                }
                            }
                        }
                    }
                }
            )(
                gr_client_wrapper(self.client, api_name)
            )

            router.get(
                path=f"{api_name}/file_fields",
                openapi_extra={
                    "responses": {
                        "200": {
                            "description": "Successful response",
                            "content": {
                                "application/json": {
                                    "schema": {
                                        "type": "object",
                                        "properties": {
                                            "inputs": {
                                                "description": "The input fields which should be url.",
                                                "type": "array",
                                                "items": {
                                                    "type": "string"
                                                }
                                            },
                                            "outputs": {
                                                "description": "The output fields which should be path.",
                                                "type": "array",
                                                "items": {
                                                    "type": "string"
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            )(
                gr_filedata_paths_wrapper(self.client, api_name)
            )

            router.get(
                path=f"{api_name}/schema",
                openapi_extra={
                    "responses": {
                        "200": {
                            "description": "Successful response",
                            "content": {
                                "application/json": {
                                    "schema": {
                                        "type": "object",
                                        "description": "The original schema of the gradio API.",
                                    }
                                }
                            }
                        }
                    }
                }
            )(
                gr_original_schema_wrapper(self.client, api_name)
            )

    @property
    def ui(self):
        return from_spaces_blocks_local_storage(
            self.client.src,
            self.client.hf_token,
        )


class LocalGradioApplication(RemoteGradioApplication):
    blocks: gr.Blocks

    def __init__(
        self,
        blocks: gr.Blocks | Any,
        **kwargs,
    ):
        self.blocks = blocks
        blocks.launch(prevent_thread_lock=True)

        super().__init__(
            src=self.blocks.local_url,
            **kwargs
        )

    @property
    def ui(self):
        return self.blocks
