from typing import Any, Literal
from gradio_client.utils import handle_file, is_http_url_like, traverse
from copy import deepcopy
DEFAULT_PARAMETER_KEY = "url"


def is_schema_filedata(obj):
    return (
        isinstance(obj, dict)
        and (properties := obj.get("properties"))
        and (meta := properties.get("meta"))
        and isinstance(meta, dict)
        and meta.get("default", {}).get("_type") == "gradio.FileData"
    )


def extract_filedata_paths(schema, prefix="") -> list[str]:
    if isinstance(schema, dict):
        if is_schema_filedata(schema):
            return [prefix]

        elif schema.get("type") == "object" and "properties" in schema:
            paths = []
            for field_name, sub_schema in schema["properties"].items():
                next_prefix = f"{prefix}.{field_name}" if prefix else field_name
                paths.extend(extract_filedata_paths(sub_schema, next_prefix))
            return paths

        elif schema.get("type") == "array" and "items" in schema:
            paths = []
            paths.extend(extract_filedata_paths(
                schema["items"], f"{prefix}[]")
            )
            return paths
        elif schema_branches := schema.get("oneOf") or schema.get("anyOf") or schema.get("allOf") or []:
            paths = []
            for sub_schema in schema_branches:
                paths.extend(extract_filedata_paths(
                    sub_schema, f"{prefix}")
                )
            return paths

    return []


def process_inputs(
    inputs: dict | list | Any,
    paths: list[str],
    prefix: str = "",
    must_http_like: bool = True,
    key=DEFAULT_PARAMETER_KEY,
):
    if prefix in paths:
        if must_http_like:
            assert is_http_url_like(
                inputs[key]), f"Expected a HTTP URL-like input, got {inputs[key]}"
        return handle_file(inputs[key])

    if isinstance(inputs, dict):
        ret = {}
        for key, value in inputs.items():
            next_prefix = f"{prefix}.{key}" if prefix else key
            ret[key] = process_inputs(
                value, paths, next_prefix, must_http_like)
        return ret
    elif isinstance(inputs, list):
        return [
            process_inputs(item, paths, f"{prefix}[]", must_http_like)
            for item in inputs
        ]
    else:
        return inputs


# No need to normalize output since we direct change the openapi schema to gradio_client output format


def __default_file_parameter(*args, **kwargs):
    return {
        "properties": {DEFAULT_PARAMETER_KEY: {"type": "string", "description": "The url of the file"}},
        "required": [DEFAULT_PARAMETER_KEY],
        "type": "object"
    }


def __default_file_return(*args, **kwargs):
    return {
        "type": "string",
        "description": "The storage of file."
    }


def normalize_parameters_schema(
    schema: dict,
):
    return traverse(
        deepcopy(schema),
        __default_file_parameter,
        is_schema_filedata,
    )


def normalize_returns_schema(
    schema: dict,
):
    return traverse(
        deepcopy(schema),
        __default_file_return,
        is_schema_filedata,
    )
