from __future__ import annotations
from typing import TYPE_CHECKING, Literal

from gradio_client import Client
from gradio_client.client import Endpoint
from gradio_client.documentation import document
from packaging import version

import gradio
from gradio.exceptions import GradioVersionIncompatibleError
from gradio.processing_utils import save_base64_to_cache, to_binary

if TYPE_CHECKING:
    from gradio.blocks import Blocks


def from_spaces_blocks_local_storage(space: str, hf_token: str | None | Literal[False]) -> Blocks:
    """
    The following lines are modified from `https://github.com/gradio-app/gradio/blob/decb5944552e951525c283e069e116c1ff6b6807/gradio/external.py#L510-L538`
    We modified the following two parts
    1. `download_file` configuration in client initialization for directly download the file from server.
    2. Let all BlockFunctions do the postprocessing.
    """
    client = Client(
        space,
        hf_token=hf_token,
        # download_files=False,
        _skip_components=False,
    )
    # We set deserialize to False to avoid downloading output files from the server.
    # Instead, we serve them as URLs using the /proxy/ endpoint directly from the server.

    if client.app_version < version.Version("4.0.0b14"):
        raise GradioVersionIncompatibleError(
            f"Gradio version 4.x cannot load spaces with versions less than 4.x ({client.app_version})."
            "Please downgrade to version 3 to load this space."
        )

    # Use end_to_end_fn here to properly upload/download all files
    predict_fns = []
    for fn_index, endpoint in client.endpoints.items():
        if not isinstance(endpoint, Endpoint):
            raise TypeError(
                f"Expected endpoint to be an Endpoint, but got {type(endpoint)}"
            )
        helper = client.new_helper(fn_index)
        if endpoint.backend_fn:
            predict_fns.append(endpoint.make_end_to_end_fn(helper))
        else:
            predict_fns.append(None)

    demo = gradio.Blocks.from_config(client.config, predict_fns, client.src)

    for _, block_function in demo.fns.items():
        block_function.postprocess = True

    # type: ignore
    return demo
